/*
  QC_Socket.cpp

  Qore Programming Language

  Copyright 2003 - 2010 David Nichols

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.
  
  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.
  
  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <qore/Qore.h>
#include <qore/intern/QC_Socket.h>
#include <qore/intern/ssl_constants.h>
#include <qore/intern/QC_Queue.h>

#include <errno.h>
#include <string.h>

qore_classid_t CID_SOCKET;

static AbstractQoreNode *doReadResult(int rc, int64 val, const char *method_name, ExceptionSink *xsink) {
   if (rc > 0)
      return new QoreBigIntNode(val);

   QoreSocket::doException(rc, method_name, xsink);
   return 0;
}

static AbstractQoreNode *doSendResult(int rc, const char *method_name, ExceptionSink *xsink) {
   if (rc == -2)
      xsink->raiseException("SOCKET-NOT-OPEN", "socket must be opened before Socket::%s() call", method_name);
   else if (rc)
      xsink->raiseErrnoException("SOCKET-SEND-ERROR", rc, "Socket::%s() failed with error code %d", method_name, rc);
   return 0;
}

static AbstractQoreNode *checkOpenResult(int rc, const char *method_name, ExceptionSink *xsink) {
   if (*xsink)
      return 0;
   if (rc == -2) {
      xsink->raiseException("SOCKET-NOT-OPEN", "socket must be open before Socket::%s() call", method_name);
      return 0;
   }
   return new QoreBigIntNode(rc);
}

static void SOCKET_constructor(QoreObject *self, const QoreListNode *params, ExceptionSink *xsink) {
   self->setPrivate(CID_SOCKET, new mySocket);
}

static void SOCKET_copy(QoreObject *self, QoreObject *old, AbstractPrivateData *obj, ExceptionSink *xsink) {
   self->setPrivate(CID_SOCKET, new mySocket);
}

// currently hardcoded to SOCK_STREAM
// opens and connects to a remote socket
// for AF_INET sockets:
// * connect("hostname:<port_number>");
// for AF_UNIX sockets:
// * connect("filename");

// Socket::connect(string $sock, int|date $timeout_ms = -1)
static AbstractQoreNode *SOCKET_connect_str_timeout(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   const QoreStringNode *p0 = HARD_QORE_STRING(params, 0);
   s->connect(p0->getBuffer(), getMsMinusOneInt(get_param(params, 1)), xsink);
   return 0;
}

// Socket::connectINET(string $host, int $port, int $timeout_ms = -1) returns nothing
// Socket::connectINET(string $host, int $port, date $timeout_ms) returns nothing
static AbstractQoreNode *SOCKET_connectINET_str_int_timeout(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   const QoreStringNode *p0 = HARD_QORE_STRING(params, 0);
   int port = HARD_QORE_INT(params, 1);
   int timeout_ms = HARD_QORE_INT(params, 2);
   s->connectINET(p0->getBuffer(), port, timeout_ms, xsink);
   return 0;
}

// Socket::connectUNIX(string $path) returns nothing
static AbstractQoreNode *SOCKET_connectUNIX(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   const QoreStringNode *p0 = HARD_QORE_STRING(params, 0);
   s->connectUNIX(p0->getBuffer(), xsink);
   return 0;
}

// currently hardcoded to SOCK_STREAM
// opens and connects to a remote socket and negotiates an SSL connection
// for AF_INET sockets:
// * connectSSL("hostname:<port_number>");
// for AF_UNIX sockets:
// * connectSSL("filename");
// Socket::connectSSL(string $sock, int|date $timeout_ms = -1)
static AbstractQoreNode *SOCKET_connectSSL_str_timeout(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   const QoreStringNode *p0 = HARD_QORE_STRING(params, 0);
   s->connectSSL(p0->getBuffer(), getMsMinusOneInt(get_param(params, 1)), xsink);
   return 0;
}

// Socket::connectINETSSL(string $host, int $port, int $timeout_ms = -1) returns nothing
// Socket::connectINETSSL(string $host, int $port, date $timeout_ms) returns nothing
static AbstractQoreNode *SOCKET_connectINETSSL_str_int_timeout(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   const QoreStringNode *p0 = HARD_QORE_STRING(params, 0);
   int port = HARD_QORE_INT(params, 1);
   int timeout_ms = HARD_QORE_INT(params, 2);
   s->connectINETSSL(p0->getBuffer(), port, timeout_ms, xsink);
   return 0;
}

// Socket::connectUNIXSSL(string $path) returns nothing
static AbstractQoreNode *SOCKET_connectUNIXSSL(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   const QoreStringNode *p0 = HARD_QORE_STRING(params, 0);
   s->connectUNIXSSL(p0->getBuffer(), xsink);
   return 0;
}

// currently hardcoded to SOCK_STREAM
// opens and binds to a local socket
// for AF_INET (tcp) sockets:
// * Socket::bind(<port_number>);
// for AF_UNIX (domain, file-based) sockets:
// * Socket::bind("filename");
// for INET (tcp) sockets
// * Socket::bind("iterface:port");

// Socket::bind(string $str, bool $reuseaddr = False) returns int
static AbstractQoreNode *SOCKET_bind_str_bool(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   const QoreStringNode *p0 = HARD_QORE_STRING(params, 0);
   bool reuseaddr = HARD_QORE_BOOL(params, 1);
   return new QoreBigIntNode(s->bind(p0->getBuffer(), reuseaddr));
}

// Socket::bind(int $port, bool $reuseaddr = False) returns int
static AbstractQoreNode *SOCKET_bind_int_bool(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   int port = HARD_QORE_INT(params, 0);
   bool reuseaddr = HARD_QORE_BOOL(params, 1);
   return new QoreBigIntNode(s->bind(port, reuseaddr));
}

// Socket::accept()
// returns a new Socket object, connection source address is in $.source
// member of new object, hostname in $.source_host
static AbstractQoreNode *SOCKET_accept(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   SocketSource source;
   mySocket *n = s->accept(&source, xsink);
   if (xsink->isEvent())
      return 0;

   // ensure that a socket object is returned (and not a subclass)
   QoreObject *ns = new QoreObject(self->getClass(CID_SOCKET), getProgram());
   ns->setPrivate(CID_SOCKET, n);
   source.setAll(ns, xsink);
      
   return ns;
}

// Socket::acceptSSL()
// accepts a new connection, negotiates an SSL connection, and returns the new socket
// the connection source string is in the "$.source" member of new object,
// hostname in "$.source_host"
static AbstractQoreNode *SOCKET_acceptSSL(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   SocketSource source;
   mySocket *n = s->acceptSSL(&source, xsink);
   if (xsink->isEvent())
      return 0;

   // ensure that a socket object is returned (and not a subclass)
   QoreObject *ns = new QoreObject(self->getClass(CID_SOCKET), getProgram());
   ns->setPrivate(CID_SOCKET, n);
   source.setAll(ns, xsink);
   
   return ns;
}

static AbstractQoreNode *SOCKET_listen(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   return checkOpenResult(s->listen(), "listen", xsink);
}

// Socket::send(binary $bin) returns int
static AbstractQoreNode *SOCKET_send_bin(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   return checkOpenResult(s->send(HARD_QORE_BINARY(params, 0)), "send", xsink);
}

// Socket::send(string $str) returns int
static AbstractQoreNode *SOCKET_send_str(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   return checkOpenResult(s->send(HARD_QORE_STRING(params, 0), xsink), "send", xsink);
}

// Socket::sendBinary(string $str) returns int
static AbstractQoreNode *SOCKET_sendBinary_str(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   const QoreStringNode *str = HARD_QORE_STRING(params, 0);
   // send without any conversions
   return checkOpenResult(s->send(str->getBuffer(), str->strlen()), "send", xsink);
}

static AbstractQoreNode *SOCKET_sendi1(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   char i = HARD_QORE_INT(params, 0);
   return checkOpenResult(s->sendi1(i), "sendi1", xsink);
}

static AbstractQoreNode *SOCKET_sendi2(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   short i = HARD_QORE_INT(params, 0);
   return checkOpenResult(s->sendi2(i), "sendi2", xsink);
}

static AbstractQoreNode *SOCKET_sendi4(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   int i = HARD_QORE_INT(params, 0);
   return checkOpenResult(s->sendi4(i), "sendi4", xsink);
}

static AbstractQoreNode *SOCKET_sendi8(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   int64 i = HARD_QORE_INT(params, 0);
   return checkOpenResult(s->sendi8(i), "sendi8", xsink);
}

static AbstractQoreNode *SOCKET_sendi2LSB(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   short i = HARD_QORE_INT(params, 0);
   return checkOpenResult(s->sendi2LSB(i), "sendi2LSB", xsink);
}

static AbstractQoreNode *SOCKET_sendi4LSB(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   int i = HARD_QORE_INT(params, 0);
   return checkOpenResult(s->sendi4LSB(i), "sendi4LSB", xsink);
}

static AbstractQoreNode *SOCKET_sendi8LSB(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   int64 i = HARD_QORE_INT(params, 0);
   return checkOpenResult(s->sendi8LSB(i), "sendi8LSB", xsink);
}

// Socket::recv(int $size = 0, softint $timeout = -1) returns string
// Socket::recv(int $size = 0, date $timeout) returns string
static AbstractQoreNode *SOCKET_recv(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   int bs = HARD_QORE_INT(params, 0);
   
   // get timeout
   int timeout = getMsMinusOneInt(get_param(params, 1));
   int rc;
   QoreStringNodeHolder msg(bs > 0 ? s->recv(bs, timeout, &rc) : s->recv(timeout, &rc));

   if (rc > 0)
      return msg.release();

   QoreSocket::doException(rc, "recv", xsink);
   return 0;
}

// Socket::recvBinary(softint $size = 0, softint $timeout = -1) returns binary
// Socket::recvBinary(softint $size = 0, date $timeout) returns binary
static AbstractQoreNode *SOCKET_recvBinary(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   int bs = HARD_QORE_INT(params, 0);

   // get timeout
   int timeout = getMsMinusOneInt(get_param(params, 1));
   int rc;
   SimpleRefHolder<BinaryNode> b(bs > 0 ? s->recvBinary(bs, timeout, &rc) : s->recvBinary(timeout, &rc));

   if (rc > 0)
      return b.release();

   QoreSocket::doException(rc, "recvBinary", xsink);
   return 0;
}

// Socket::recvi1(softint $timeout = -1)
// Socket::recvi1(date $timeout)
static AbstractQoreNode *SOCKET_recvi1(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   // get timeout
   int timeout = getMsMinusOneInt(get_param(params, 0));
   char b;
   int rc = s->recvi1(timeout, &b);
   return doReadResult(rc, (int64)b, "recvi1", xsink);
}

static AbstractQoreNode *SOCKET_recvi2(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   // get timeout
   int timeout = getMsMinusOneInt(get_param(params, 0));
   short b;
   int rc = s->recvi2(timeout, &b);
   return doReadResult(rc, (int64)b, "recvi2", xsink);
}

static AbstractQoreNode *SOCKET_recvi4(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   // get timeout
   int timeout = getMsMinusOneInt(get_param(params, 0));
   int b;
   int rc = s->recvi4(timeout, &b);
   return doReadResult(rc, (int64)b, "recvi4", xsink);
}

static AbstractQoreNode *SOCKET_recvi8(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   // get timeout
   int timeout = getMsMinusOneInt(get_param(params, 0));
   int64 b;
   int rc = s->recvi8(timeout, &b);
   return doReadResult(rc, b, "recvi8", xsink);
}

static AbstractQoreNode *SOCKET_recvi2LSB(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   // get timeout
   int timeout = getMsMinusOneInt(get_param(params, 0));
   short b;
   int rc = s->recvi2LSB(timeout, &b);
   return doReadResult(rc, (int64)b, "recvi2LSB", xsink);
}

static AbstractQoreNode *SOCKET_recvi4LSB(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   // get timeout
   int timeout = getMsMinusOneInt(get_param(params, 0));
   int b;
   int rc = s->recvi4LSB(timeout, &b);
   return doReadResult(rc, (int64)b, "recvi4LSB", xsink);
}

static AbstractQoreNode *SOCKET_recvi8LSB(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   // get timeout
   int timeout = getMsMinusOneInt(get_param(params, 0));
   int64 b;
   int rc = s->recvi8LSB(timeout, &b);
   return doReadResult(rc, b, "recvi8LSB", xsink);
}

static AbstractQoreNode *SOCKET_recvu1(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   // get timeout
   int timeout = getMsMinusOneInt(get_param(params, 0));   
   unsigned char b;
   int rc = s->recvu1(timeout, &b);
   return doReadResult(rc, (int64)b, "recvu1", xsink);
}

static AbstractQoreNode *SOCKET_recvu2(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   // get timeout
   int timeout = getMsMinusOneInt(get_param(params, 0));
   unsigned short b;
   int rc = s->recvu2(timeout, &b);
   return doReadResult(rc, (int64)b, "recvu2", xsink);
}

static AbstractQoreNode *SOCKET_recvu4(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   // get timeout
   int timeout = getMsMinusOneInt(get_param(params, 0));   
   unsigned int b;
   int rc = s->recvu4(timeout, &b);
   return doReadResult(rc, (int64)b, "recvu4", xsink);
}

static AbstractQoreNode *SOCKET_recvu2LSB(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   // get timeout
   int timeout = getMsMinusOneInt(get_param(params, 0));
   unsigned short b;
   int rc = s->recvu2LSB(timeout, &b);
   return doReadResult(rc, (int64)b, "recvu2LSB", xsink);
}

static AbstractQoreNode *SOCKET_recvu4LSB(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   // get timeout
   int timeout = getMsMinusOneInt(get_param(params, 0));
   
   unsigned b;
   int rc = s->recvu4LSB(timeout, &b);
   return doReadResult(rc, (int64)b, "recvu4LSB", xsink);
}

// Socket::sendHTTPMessage(string $method, string $path, string $http_version, hash $headers, data $data = binary()) returns int
static AbstractQoreNode *SOCKET_sendHTTPMessage(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   const char *method = HARD_QORE_STRING(params, 0)->getBuffer();
   const char *path = HARD_QORE_STRING(params, 1)->getBuffer();
   const char *http_version = HARD_QORE_STRING(params, 2)->getBuffer();
   const QoreHashNode *headers = HARD_QORE_HASH(params, 3);

   // see if there is data to send as well
   const AbstractQoreNode *p4 = get_param(params, 4);
   const void *ptr = 0;
   int size = 0;

   if (p4->getType() == NT_STRING) {
      const QoreStringNode *str = reinterpret_cast<const QoreStringNode *>(p4);
      ptr = str->getBuffer();
      size = str->strlen();
   }
   else {
      assert(p4->getType() == NT_BINARY);
      const BinaryNode *b = reinterpret_cast<const BinaryNode *>(p4);
      ptr = b->getPtr();
      size = b->size();
   }

   int rc = s->sendHTTPMessage(method, path, http_version, headers, ptr, size);
   return doSendResult(rc, "sendHTTPMessage", xsink);
}

// Socket::sendHTTPResponse(softint $status_code, string $desc, string $http_version, hash $headers, data $data = binary()) returns nothing
static AbstractQoreNode *SOCKET_sendHTTPResponse(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   int status_code = HARD_QORE_INT(params, 0);

   if (status_code < 100 || status_code >= 600) {
      xsink->raiseException("SOCKET-SENDHTTPRESPONSE-PARAMETER-ERROR", "expecting valid HTTP status code between 100 and 599 as first parameter of Socket::sendHTTPResponse() call, got value %d instead", status_code);
      return 0;
   }

   const char *status_desc = HARD_QORE_STRING(params, 1)->getBuffer();
   const char *http_version = HARD_QORE_STRING(params, 2)->getBuffer();
   const QoreHashNode *headers = HARD_QORE_HASH(params, 3);

   // see if there is data to send as well
   const AbstractQoreNode *p4 = get_param(params, 4);
   const void *ptr = 0;
   int size = 0;

   if (p4->getType() == NT_STRING) {
      const QoreStringNode *str = reinterpret_cast<const QoreStringNode *>(p4);
      ptr = str->getBuffer();
      size = str->strlen();
   }
   else {
      assert(p4->getType() == NT_BINARY);
      const BinaryNode *b = reinterpret_cast<const BinaryNode *>(p4);
      ptr = b->getPtr();
      size = b->size();
   }

   int rc = s->sendHTTPResponse(status_code, status_desc, http_version, headers, ptr, size);
   return doSendResult(rc, "sendHTTPResponse", xsink);
}

// Socket::readHTTPHeader(softint $timeout_ms = -1) returns string|hash
// Socket::readHTTPHeader(date $timeout_ms) returns string|hash
static AbstractQoreNode *SOCKET_readHTTPHeader(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   int timeout = getMsMinusOneInt(get_param(params, 0));
   int rc;

   // when rc = -3 it's a timeout, but rv will be NULL anyway, so we do nothing
   AbstractQoreNode *rv = s->readHTTPHeader(timeout, &rc);
      
   if (rc <= 0)
      QoreSocket::doException(rc, "readHTTPHeader", xsink);

   return rv;
}

// Socket::readHTTPChunkedBody(softint $timeout_ms = -1) returns hash
// Socket::readHTTPChunkedBody(date $timeout_ms) returns hash
static AbstractQoreNode *SOCKET_readHTTPChunkedBody(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   int timeout = getMsMinusOneInt(get_param(params, 0));

   // when rc = -3 it's a timeout, but rv will be NULL anyway, so we do nothing
   return s->readHTTPChunkedBody(timeout, xsink);
}

// Socket::readHTTPChunkedBodyBinary(softint $timeout_ms = -1) returns hash
// Socket::readHTTPChunkedBodyBinary(date $timeout_ms) returns hash
static AbstractQoreNode *SOCKET_readHTTPChunkedBodyBinary(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   int timeout = getMsMinusOneInt(get_param(params, 0));

   // when rc = -3 it's a timeout, but rv will be NULL anyway, so we do nothing
   return s->readHTTPChunkedBodyBinary(timeout, xsink);
}

static AbstractQoreNode *SOCKET_close(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   return new QoreBigIntNode(s->close());
}

static AbstractQoreNode *SOCKET_shutdown(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   return new QoreBigIntNode(s->shutdown());
}

static AbstractQoreNode *SOCKET_shutdownSSL(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   s->shutdownSSL(xsink);
   return 0;
}

static AbstractQoreNode *SOCKET_getPort(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   return new QoreBigIntNode(s->getPort());
}

static AbstractQoreNode *SOCKET_getSocket(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   return new QoreBigIntNode(s->getSocket());
}

// Socket::setSendTimeout(softint $timeout_ms) returns int
static AbstractQoreNode *SOCKET_setSendTimeout_int(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   int timeout_ms = HARD_QORE_INT(params, 0);
   return new QoreBigIntNode(s->setSendTimeout(timeout_ms));
}

// Socket::setSendTimeout(date $timeout_ms) returns int
static AbstractQoreNode *SOCKET_setSendTimeout_date(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   int timeout_ms = HARD_QORE_DATE(params, 0)->getRelativeMilliseconds();
   return new QoreBigIntNode(s->setSendTimeout(timeout_ms));
}

// Socket::setRecvTimeout(softint $timeout_ms) returns int
static AbstractQoreNode *SOCKET_setRecvTimeout_int(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   int timeout_ms = HARD_QORE_INT(params, 0);
   return new QoreBigIntNode(s->setRecvTimeout(timeout_ms));
}

// Socket::setRecvTimeout(date $timeout_ms) returns int
static AbstractQoreNode *SOCKET_setRecvTimeout_date(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   int timeout_ms = HARD_QORE_DATE(params, 0)->getRelativeMilliseconds();
   return new QoreBigIntNode(s->setRecvTimeout(timeout_ms));
}

static AbstractQoreNode *SOCKET_getSendTimeout(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   return new QoreBigIntNode(s->getSendTimeout());
}

static AbstractQoreNode *SOCKET_getRecvTimeout(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   return new QoreBigIntNode(s->getRecvTimeout());
}

static AbstractQoreNode *SOCKET_getCharset(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   return new QoreStringNode(s->getEncoding()->getCode());
}

static AbstractQoreNode *SOCKET_setCharset(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   const QoreStringNode *p0 = HARD_QORE_STRING(params, 0);
   s->setEncoding(QEM.findCreate(p0));
   return 0; 
}

static AbstractQoreNode *SOCKET_isDataAvailable(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   return get_bool_node(s->isDataAvailable(getMsZeroInt(get_param(params, 0))));
}

static AbstractQoreNode *SOCKET_isWriteFinished(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   return get_bool_node(s->isWriteFinished(getMsZeroInt(get_param(params, 0))));
}

static AbstractQoreNode *SOCKET_isOpen(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   return get_bool_node(s->isOpen());
}

// Socket::getSSLCipherName() returns string|nothing
static AbstractQoreNode *SOCKET_getSSLCipherName(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   const char *str = s->getSSLCipherName();
   return str ? new QoreStringNode(str) : 0;
}

// Socket::getSSLCipherVersion() returns string|nothing
static AbstractQoreNode *SOCKET_getSSLCipherVersion(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   const char *str = s->getSSLCipherVersion();
   return str ? new QoreStringNode(str) : 0;
}

static AbstractQoreNode *SOCKET_isSecure(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   return get_bool_node(s->isSecure());
}

// Socket::verifyPeerCertificate() returns string|nothing
static AbstractQoreNode *SOCKET_verifyPeerCertificate(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   const char *c = getSSLCVCode(s->verifyPeerCertificate());
   return c ? new QoreStringNode(c) : 0;
}

// Socket::setCertificate(SSLCertificate $cert) returns nothing
static AbstractQoreNode *SOCKET_setCertificate_cert(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   HARD_QORE_OBJ_DATA(cert, QoreSSLCertificate, params, 0, CID_SSLCERTIFICATE, "Socket::setCertificate()", "SSLCertificate", xsink);
   if (*xsink)
      return 0;

   // pass reference from QoreObject::getReferencedPrivateData() to mySocket::setCertificate()
   s->setCertificate(cert);
   return 0;
}

// Socket::setCertificate(string $cert_pem) returns nothing
static AbstractQoreNode *SOCKET_setCertificate_string(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   const QoreStringNode *p0 = HARD_QORE_STRING(params, 0);
   SimpleRefHolder<QoreSSLCertificate> cert(new QoreSSLCertificate(p0, xsink));
   if (*xsink)
      return 0;

   // mySocket::setCertificate() takes over ownership of certificate reference
   s->setCertificate(cert.release());
   return 0;   
}

// Socket::setCertificate(binary $cert_der) returns nothing
static AbstractQoreNode *SOCKET_setCertificate_bin(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   const BinaryNode *p0 = HARD_QORE_BINARY(params, 0);
   SimpleRefHolder<QoreSSLCertificate> cert(new QoreSSLCertificate(p0, xsink));
   if (*xsink)
      return 0;

   // mySocket::setCertificate() takes over ownership of certificate reference
   s->setCertificate(cert.release());
   return 0;   
}

// Socket::setPrivateKey(SSLPrivateKey $key)
static AbstractQoreNode *SOCKET_setPrivateKey_key(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   HARD_QORE_OBJ_DATA(key, QoreSSLPrivateKey, params, 0, CID_SSLPRIVATEKEY, "Socket::setPrivateKey()", "SSLPrivateKey", xsink);
   if (*xsink)
      return 0;

   // pass reference from QoreObject::getReferencedPrivateData() to mySocket::setPrivateKey()
   s->setPrivateKey(key);
   return 0;
}

// Socket::setPrivateKey(string $key_pem)
static AbstractQoreNode *SOCKET_setPrivateKey_str(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   const QoreStringNode *p0 = HARD_QORE_STRING(params, 0);
   SimpleRefHolder<QoreSSLPrivateKey> key(new QoreSSLPrivateKey(p0, 0, xsink));
   if (*xsink)
      return 0;

   s->setPrivateKey(key.release());
   return 0;
}

// Socket::setPrivateKey(string $key_pem, string $pass)
static AbstractQoreNode *SOCKET_setPrivateKey_str_str(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   const QoreStringNode *p0 = HARD_QORE_STRING(params, 0);
   const QoreStringNode *p1 = HARD_QORE_STRING(params, 1);
   SimpleRefHolder<QoreSSLPrivateKey> key(new QoreSSLPrivateKey(p0, p1->getBuffer(), xsink));
   if (*xsink)
      return 0;

   s->setPrivateKey(key.release());
   return 0;
}

// Socket::setPrivateKey(binary $key_der)
static AbstractQoreNode *SOCKET_setPrivateKey_bin(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   const BinaryNode *p0 = HARD_QORE_BINARY(params, 0);
   SimpleRefHolder<QoreSSLPrivateKey> key(new QoreSSLPrivateKey(p0, xsink));
   if (*xsink)
      return 0;

   s->setPrivateKey(key.release());
   return 0;
}

static AbstractQoreNode *SOCKET_setEventQueue_nothing(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   s->setEventQueue(0, xsink);
   return 0;
}

static AbstractQoreNode *SOCKET_setEventQueue_queue(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   HARD_QORE_OBJ_DATA(q, Queue, params, 0, CID_QUEUE, "Queue", "Socket::setEventQueue", xsink);
   if (*xsink)
      return 0;
   // pass reference from QoreObject::getReferencedPrivateData() to function
   s->setEventQueue(q, xsink);
   return 0;
}

static AbstractQoreNode *SOCKET_setNoDelay(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   return new QoreBigIntNode(s->setNoDelay(HARD_QORE_BOOL(params, 0)));
}

static AbstractQoreNode *SOCKET_getNoDelay(QoreObject *self, mySocket *s, const QoreListNode *params, ExceptionSink *xsink) {
   return get_bool_node(s->getNoDelay());
}

QoreClass *initSocketClass(QoreClass *SSLCert, QoreClass *SSLPrivKey) {
   QORE_TRACE("initSocketClass()");

   assert(QC_QUEUE);

   QoreClass *QC_SOCKET = new QoreClass("Socket", QDOM_NETWORK);
   CID_SOCKET = QC_SOCKET->getID();

   // register public members
   QC_SOCKET->addPublicMember("source", anyTypeInfo);
   QC_SOCKET->addPublicMember("source_host", anyTypeInfo);
   // unset the public member flag for backwards-compatibility
   // this allows older code where a user class inherits this class to still be able
   // to be run in new Qore >= 0.8
   QC_SOCKET->unsetPublicMemberFlag();
   
   QC_SOCKET->setConstructorExtended(SOCKET_constructor);

   QC_SOCKET->setCopy(SOCKET_copy);

   // Socket::connect(string $sock, int $timeout_ms = -1)
   QC_SOCKET->addMethodExtended("connect",                   (q_method_t)SOCKET_connect_str_timeout, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 2, stringTypeInfo, QORE_PARAM_NO_ARG, softBigIntTypeInfo, new QoreBigIntNode(-1));
   // Socket::connect(string $sock, date $timeout_ms)
   QC_SOCKET->addMethodExtended("connect",                   (q_method_t)SOCKET_connect_str_timeout, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 2, stringTypeInfo, QORE_PARAM_NO_ARG, dateTypeInfo, QORE_PARAM_NO_ARG);

   // Socket::connectINET(string $host, int $port, int $timeout_ms = -1) returns nothing
   QC_SOCKET->addMethodExtended("connectINET",               (q_method_t)SOCKET_connectINET_str_int_timeout, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 3, stringTypeInfo, QORE_PARAM_NO_ARG, softBigIntTypeInfo, QORE_PARAM_NO_ARG, softBigIntTypeInfo, new QoreBigIntNode(-1));
   // Socket::connectINET(string $host, int $port, date $timeout_ms) returns nothing
   QC_SOCKET->addMethodExtended("connectINET",               (q_method_t)SOCKET_connectINET_str_int_timeout, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 3, stringTypeInfo, QORE_PARAM_NO_ARG, softBigIntTypeInfo, QORE_PARAM_NO_ARG, dateTypeInfo, QORE_PARAM_NO_ARG);

   // Socket::connectUNIX(string $path) returns nothing
   QC_SOCKET->addMethodExtended("connectUNIX",               (q_method_t)SOCKET_connectUNIX, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, stringTypeInfo, QORE_PARAM_NO_ARG);

   // Socket::connectSSL(string $sock, int $timeout_ms = -1)
   QC_SOCKET->addMethodExtended("connectSSL",                (q_method_t)SOCKET_connectSSL_str_timeout, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 2, stringTypeInfo, QORE_PARAM_NO_ARG, softBigIntTypeInfo, new QoreBigIntNode(-1));
   // Socket::connectSSL(string $sock, date $timeout_ms)
   QC_SOCKET->addMethodExtended("connectSSL",                (q_method_t)SOCKET_connectSSL_str_timeout, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 2, stringTypeInfo, QORE_PARAM_NO_ARG, dateTypeInfo, QORE_PARAM_NO_ARG);

   // Socket::connectINETSSL(string $host, int $port, int $timeout_ms = -1) returns nothing
   QC_SOCKET->addMethodExtended("connectINETSSL",            (q_method_t)SOCKET_connectINETSSL_str_int_timeout, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 3, stringTypeInfo, QORE_PARAM_NO_ARG, softBigIntTypeInfo, QORE_PARAM_NO_ARG, softBigIntTypeInfo, new QoreBigIntNode(-1));
   // Socket::connectINETSSL(string $host, int $port, date $timeout_ms) returns nothing
   QC_SOCKET->addMethodExtended("connectINETSSL",            (q_method_t)SOCKET_connectINETSSL_str_int_timeout, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 3, stringTypeInfo, QORE_PARAM_NO_ARG, softBigIntTypeInfo, QORE_PARAM_NO_ARG, dateTypeInfo, QORE_PARAM_NO_ARG);

   // Socket::connectUNIXSSL(string $path) returns nothing
   QC_SOCKET->addMethodExtended("connectUNIXSSL",            (q_method_t)SOCKET_connectUNIXSSL, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, stringTypeInfo, QORE_PARAM_NO_ARG);

   // Socket::bind(string $str, bool $reuseaddr = False) returns int
   QC_SOCKET->addMethodExtended("bind",                      (q_method_t)SOCKET_bind_str_bool, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 2, stringTypeInfo, QORE_PARAM_NO_ARG, boolTypeInfo, &False);

   // Socket::bind(int $port, bool $reuseaddr = False) returns int
   QC_SOCKET->addMethodExtended("bind",                      (q_method_t)SOCKET_bind_int_bool, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 2, bigIntTypeInfo, QORE_PARAM_NO_ARG, boolTypeInfo, &False);

   QC_SOCKET->addMethodExtended("accept",                    (q_method_t)SOCKET_accept, false, QC_NO_FLAGS, QDOM_DEFAULT, QC_SOCKET->getTypeInfo());
   QC_SOCKET->addMethodExtended("acceptSSL",                 (q_method_t)SOCKET_acceptSSL, false, QC_NO_FLAGS, QDOM_DEFAULT, QC_SOCKET->getTypeInfo());

   QC_SOCKET->addMethodExtended("listen",                    (q_method_t)SOCKET_listen, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo);

   // Socket::send(binary $bin) returns int
   QC_SOCKET->addMethodExtended("send",                      (q_method_t)SOCKET_send_bin, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, binaryTypeInfo, QORE_PARAM_NO_ARG);

   // Socket::send(string $str) returns int
   QC_SOCKET->addMethodExtended("send",                      (q_method_t)SOCKET_send_str, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, stringTypeInfo, QORE_PARAM_NO_ARG);

   // Socket::sendBinary(binary $bin) returns int
   QC_SOCKET->addMethodExtended("sendBinary",                (q_method_t)SOCKET_send_bin, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, binaryTypeInfo, QORE_PARAM_NO_ARG);

   // Socket::sendBinary(string $str) returns int
   QC_SOCKET->addMethodExtended("sendBinary",                (q_method_t)SOCKET_sendBinary_str, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, stringTypeInfo, QORE_PARAM_NO_ARG);

   QC_SOCKET->addMethodExtended("sendi1",                    (q_method_t)SOCKET_sendi1, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, softBigIntTypeInfo, zero());
   QC_SOCKET->addMethodExtended("sendi2",                    (q_method_t)SOCKET_sendi2, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, softBigIntTypeInfo, zero());
   QC_SOCKET->addMethodExtended("sendi4",                    (q_method_t)SOCKET_sendi4, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, softBigIntTypeInfo, zero());
   QC_SOCKET->addMethodExtended("sendi8",                    (q_method_t)SOCKET_sendi8, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, softBigIntTypeInfo, zero());
   QC_SOCKET->addMethodExtended("sendi2LSB",                 (q_method_t)SOCKET_sendi2LSB, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, softBigIntTypeInfo, zero());
   QC_SOCKET->addMethodExtended("sendi4LSB",                 (q_method_t)SOCKET_sendi4LSB, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, softBigIntTypeInfo, zero());
   QC_SOCKET->addMethodExtended("sendi8LSB",                 (q_method_t)SOCKET_sendi8LSB, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, softBigIntTypeInfo, zero());

   // Socket::recv(softint $size = 0, softint $timeout = -1) returns string
   QC_SOCKET->addMethodExtended("recv",                      (q_method_t)SOCKET_recv, false, QC_NO_FLAGS, QDOM_DEFAULT, stringTypeInfo, 2, softBigIntTypeInfo, zero(), softBigIntTypeInfo, new QoreBigIntNode(-1));
   // Socket::recv(softint $size = 0, date $timeout) returns string
   QC_SOCKET->addMethodExtended("recv",                      (q_method_t)SOCKET_recv, false, QC_NO_FLAGS, QDOM_DEFAULT, stringTypeInfo, 2, softBigIntTypeInfo, zero(), dateTypeInfo, QORE_PARAM_NO_ARG);

   // Socket::recvBinary(softint $size = 0, softint $timeout = -1) returns binary
   QC_SOCKET->addMethodExtended("recvBinary",                (q_method_t)SOCKET_recvBinary, false, QC_NO_FLAGS, QDOM_DEFAULT, binaryTypeInfo, 2, softBigIntTypeInfo, zero(), softBigIntTypeInfo, new QoreBigIntNode(-1));
   // Socket::recvBinary(softint $size = 0, date $timeout) returns binary
   QC_SOCKET->addMethodExtended("recvBinary",                (q_method_t)SOCKET_recvBinary, false, QC_NO_FLAGS, QDOM_DEFAULT, stringTypeInfo, 2, softBigIntTypeInfo, zero(), dateTypeInfo, QORE_PARAM_NO_ARG);

   // Socket::recvi1(softint $timeout = -1) returns int
   QC_SOCKET->addMethodExtended("recvi1",                    (q_method_t)SOCKET_recvi1, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, softBigIntTypeInfo, new QoreBigIntNode(-1));
   // Socket::recvi1(date $timeout) returns int
   QC_SOCKET->addMethodExtended("recvi1",                    (q_method_t)SOCKET_recvi1, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, dateTypeInfo, QORE_PARAM_NO_ARG);

   // Socket::recvi2(softint $timeout = -1) returns int
   QC_SOCKET->addMethodExtended("recvi2",                    (q_method_t)SOCKET_recvi2, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, softBigIntTypeInfo, new QoreBigIntNode(-1));
   // Socket::recvi2(date $timeout) returns int
   QC_SOCKET->addMethodExtended("recvi2",                    (q_method_t)SOCKET_recvi2, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, dateTypeInfo, QORE_PARAM_NO_ARG);

   // Socket::recvi4(softint $timeout = -1) returns int
   QC_SOCKET->addMethodExtended("recvi4",                    (q_method_t)SOCKET_recvi4, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, softBigIntTypeInfo, new QoreBigIntNode(-1));
   // Socket::recvi4(date $timeout) returns int
   QC_SOCKET->addMethodExtended("recvi4",                    (q_method_t)SOCKET_recvi4, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, dateTypeInfo, QORE_PARAM_NO_ARG);

   // Socket::recvi8(softint $timeout = -1) returns int
   QC_SOCKET->addMethodExtended("recvi8",                    (q_method_t)SOCKET_recvi8, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, softBigIntTypeInfo, new QoreBigIntNode(-1));
   // Socket::recvi8(date $timeout) returns int
   QC_SOCKET->addMethodExtended("recvi8",                    (q_method_t)SOCKET_recvi8, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, dateTypeInfo, QORE_PARAM_NO_ARG);

   // Socket::recvi2LSB(softint $timeout = -1) returns int
   QC_SOCKET->addMethodExtended("recvi2LSB",                 (q_method_t)SOCKET_recvi2LSB, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, softBigIntTypeInfo, new QoreBigIntNode(-1));
   // Socket::recvi2LSB(date $timeout) returns int
   QC_SOCKET->addMethodExtended("recvi2LSB",                 (q_method_t)SOCKET_recvi2LSB, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, dateTypeInfo, QORE_PARAM_NO_ARG);

   // Socket::recvi4LSB(softint $timeout = -1) returns int
   QC_SOCKET->addMethodExtended("recvi4LSB",                 (q_method_t)SOCKET_recvi4LSB, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, softBigIntTypeInfo, new QoreBigIntNode(-1));
   // Socket::recvi4LSB(date $timeout) returns int
   QC_SOCKET->addMethodExtended("recvi4LSB",                 (q_method_t)SOCKET_recvi4LSB, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, dateTypeInfo, QORE_PARAM_NO_ARG);

   // Socket::recvi8LSB(softint $timeout = -1) returns int
   QC_SOCKET->addMethodExtended("recvi8LSB",                 (q_method_t)SOCKET_recvi8LSB, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, softBigIntTypeInfo, new QoreBigIntNode(-1));
   // Socket::recvi8LSB(date $timeout) returns int
   QC_SOCKET->addMethodExtended("recvi8LSB",                 (q_method_t)SOCKET_recvi8LSB, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, dateTypeInfo, QORE_PARAM_NO_ARG);

   // Socket::recvu1(softint $timeout = -1) returns int
   QC_SOCKET->addMethodExtended("recvu1",                    (q_method_t)SOCKET_recvu1, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, softBigIntTypeInfo, new QoreBigIntNode(-1));
   // Socket::recvu1(date $timeout) returns int
   QC_SOCKET->addMethodExtended("recvu1",                    (q_method_t)SOCKET_recvu1, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, dateTypeInfo, QORE_PARAM_NO_ARG);

   // Socket::recvu2(softint $timeout = -1) returns int
   QC_SOCKET->addMethodExtended("recvu2",                    (q_method_t)SOCKET_recvu2, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, softBigIntTypeInfo, new QoreBigIntNode(-1));
   // Socket::recvu2(date $timeout) returns int
   QC_SOCKET->addMethodExtended("recvu2",                    (q_method_t)SOCKET_recvu2, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, dateTypeInfo, QORE_PARAM_NO_ARG);

   // Socket::recvu4(softint $timeout = -1) returns int
   QC_SOCKET->addMethodExtended("recvu4",                    (q_method_t)SOCKET_recvu4, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, softBigIntTypeInfo, new QoreBigIntNode(-1));
   // Socket::recvu4(date $timeout) returns int
   QC_SOCKET->addMethodExtended("recvu4",                    (q_method_t)SOCKET_recvu4, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, dateTypeInfo, QORE_PARAM_NO_ARG);

   // Socket::recvu2LSB(softint $timeout = -1) returns int
   QC_SOCKET->addMethodExtended("recvu2LSB",                 (q_method_t)SOCKET_recvu2LSB, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, softBigIntTypeInfo, new QoreBigIntNode(-1));
   // Socket::recvu2LSB(date $timeout) returns int
   QC_SOCKET->addMethodExtended("recvu2LSB",                 (q_method_t)SOCKET_recvu2LSB, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, dateTypeInfo, QORE_PARAM_NO_ARG);

   // Socket::recvu4LSB(softint $timeout = -1) returns int
   QC_SOCKET->addMethodExtended("recvu4LSB",                 (q_method_t)SOCKET_recvu4LSB, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, softBigIntTypeInfo, new QoreBigIntNode(-1));
   // Socket::recvu4LSB(date $timeout) returns int
   QC_SOCKET->addMethodExtended("recvu4LSB",                 (q_method_t)SOCKET_recvu4LSB, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, dateTypeInfo, QORE_PARAM_NO_ARG);

   // Socket::sendHTTPMessage(string $method, string $path, string $http_version, hash $headers, data $data = binary()) returns nothing
   QC_SOCKET->addMethodExtended("sendHTTPMessage",           (q_method_t)SOCKET_sendHTTPMessage, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 5, stringTypeInfo, QORE_PARAM_NO_ARG, stringTypeInfo, QORE_PARAM_NO_ARG, stringTypeInfo, QORE_PARAM_NO_ARG, hashTypeInfo, QORE_PARAM_NO_ARG, dataTypeInfo, new BinaryNode);

   // Socket::sendHTTPResponse(softint $status_code, string $desc, string $http_version, hash $headers, data $data = binary()) returns nothing
   QC_SOCKET->addMethodExtended("sendHTTPResponse",          (q_method_t)SOCKET_sendHTTPResponse, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 5, softBigIntTypeInfo, QORE_PARAM_NO_ARG, stringTypeInfo, QORE_PARAM_NO_ARG, stringTypeInfo, QORE_PARAM_NO_ARG, hashTypeInfo, QORE_PARAM_NO_ARG, dataTypeInfo, new BinaryNode);

   // Socket::readHTTPHeader(softint $timeout_ms = -1) returns string|hash
   QC_SOCKET->addMethodExtended("readHTTPHeader",            (q_method_t)SOCKET_readHTTPHeader, false, QC_NO_FLAGS, QDOM_DEFAULT, 0, 1, softBigIntTypeInfo, new QoreBigIntNode(-1));
   // Socket::readHTTPHeader(date $timeout_ms) returns string|hash
   QC_SOCKET->addMethodExtended("readHTTPHeader",            (q_method_t)SOCKET_readHTTPHeader, false, QC_NO_FLAGS, QDOM_DEFAULT, 0, 1, dateTypeInfo, QORE_PARAM_NO_ARG);

   // Socket::readHTTPChunkedBody(softint $timeout_ms = -1) returns hash
   QC_SOCKET->addMethodExtended("readHTTPChunkedBody",       (q_method_t)SOCKET_readHTTPChunkedBody, false, QC_NO_FLAGS, QDOM_DEFAULT, hashTypeInfo, 1, softBigIntTypeInfo, new QoreBigIntNode(-1));
   // Socket::readHTTPChunkedBody(date $timeout_ms) returns hash
   QC_SOCKET->addMethodExtended("readHTTPChunkedBody",       (q_method_t)SOCKET_readHTTPChunkedBody, false, QC_NO_FLAGS, QDOM_DEFAULT, hashTypeInfo, 1, dateTypeInfo, QORE_PARAM_NO_ARG);

   // Socket::readHTTPChunkedBodyBinary(softint $timeout_ms = -1) returns hash
   QC_SOCKET->addMethodExtended("readHTTPChunkedBodyBinary", (q_method_t)SOCKET_readHTTPChunkedBodyBinary, false, QC_NO_FLAGS, QDOM_DEFAULT, hashTypeInfo, 1, softBigIntTypeInfo, new QoreBigIntNode(-1));
   // Socket::readHTTPChunkedBodyBinary(date $timeout_ms) returns hash
   QC_SOCKET->addMethodExtended("readHTTPChunkedBodyBinary", (q_method_t)SOCKET_readHTTPChunkedBodyBinary, false, QC_NO_FLAGS, QDOM_DEFAULT, hashTypeInfo, 1, dateTypeInfo, QORE_PARAM_NO_ARG);

   QC_SOCKET->addMethodExtended("close",                     (q_method_t)SOCKET_close, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo);

   QC_SOCKET->addMethodExtended("shutdown",                  (q_method_t)SOCKET_shutdown, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo);

   QC_SOCKET->addMethodExtended("shutdownSSL",               (q_method_t)SOCKET_shutdownSSL, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);

   QC_SOCKET->addMethodExtended("getPort",                   (q_method_t)SOCKET_getPort, false, QC_RET_VALUE_ONLY, QDOM_DEFAULT, bigIntTypeInfo);

   QC_SOCKET->addMethodExtended("getSocket",                 (q_method_t)SOCKET_getSocket, false, QC_RET_VALUE_ONLY, QDOM_DEFAULT, bigIntTypeInfo);

   // Socket::setSendTimeout(softint $timeout_ms) returns int
   QC_SOCKET->addMethodExtended("setSendTimeout",            (q_method_t)SOCKET_setSendTimeout_int, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, softBigIntTypeInfo, QORE_PARAM_NO_ARG);
   // Socket::setSendTimeout(date $timeout_ms) returns int
   QC_SOCKET->addMethodExtended("setSendTimeout",            (q_method_t)SOCKET_setSendTimeout_date, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, dateTypeInfo, QORE_PARAM_NO_ARG);

   // Socket::setRecvTimeout(softint $timeout_ms) returns int
   QC_SOCKET->addMethodExtended("setRecvTimeout",            (q_method_t)SOCKET_setRecvTimeout_int, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, softBigIntTypeInfo, QORE_PARAM_NO_ARG);
   // Socket::setRecvTimeout(date $timeout_ms) returns int
   QC_SOCKET->addMethodExtended("setRecvTimeout",            (q_method_t)SOCKET_setRecvTimeout_date, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, dateTypeInfo, QORE_PARAM_NO_ARG);

   QC_SOCKET->addMethodExtended("getSendTimeout",            (q_method_t)SOCKET_getSendTimeout, false, QC_RET_VALUE_ONLY, QDOM_DEFAULT, bigIntTypeInfo);

   QC_SOCKET->addMethodExtended("getRecvTimeout",            (q_method_t)SOCKET_getRecvTimeout, false, QC_RET_VALUE_ONLY, QDOM_DEFAULT, bigIntTypeInfo);

   QC_SOCKET->addMethodExtended("getCharset",                (q_method_t)SOCKET_getCharset, false, QC_RET_VALUE_ONLY, QDOM_DEFAULT, stringTypeInfo);

   QC_SOCKET->addMethodExtended("setCharset",                (q_method_t)SOCKET_setCharset, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, stringTypeInfo, QORE_PARAM_NO_ARG);

   // Socket::isDataAvailable(softint $timeout_ms = 0)
   QC_SOCKET->addMethodExtended("isDataAvailable",           (q_method_t)SOCKET_isDataAvailable, false, QC_NO_FLAGS, QDOM_DEFAULT, boolTypeInfo, 1, softBigIntTypeInfo, zero());
   // Socket::isDataAvailable(date $timeout_ms)
   QC_SOCKET->addMethodExtended("isDataAvailable",           (q_method_t)SOCKET_isDataAvailable, false, QC_NO_FLAGS, QDOM_DEFAULT, boolTypeInfo, 1, dateTypeInfo, QORE_PARAM_NO_ARG);

   // Socket::isWriteFinished(softint $timeout_ms = 0)
   QC_SOCKET->addMethodExtended("isWriteFinished",           (q_method_t)SOCKET_isWriteFinished, false, QC_NO_FLAGS, QDOM_DEFAULT, boolTypeInfo, 1, softBigIntTypeInfo, zero());
   // Socket::isWriteFinished(date $timeout_ms)
   QC_SOCKET->addMethodExtended("isWriteFinished",           (q_method_t)SOCKET_isWriteFinished, false, QC_NO_FLAGS, QDOM_DEFAULT, boolTypeInfo, 1, dateTypeInfo, QORE_PARAM_NO_ARG);

   // Socket::getSSLCipherName() returns string|nothing
   QC_SOCKET->addMethodExtended("getSSLCipherName",          (q_method_t)SOCKET_getSSLCipherName, false, QC_RET_VALUE_ONLY);

   // Socket::getSSLCipherVersion() returns string|nothing
   QC_SOCKET->addMethodExtended("getSSLCipherVersion",       (q_method_t)SOCKET_getSSLCipherVersion, false, QC_RET_VALUE_ONLY);

   QC_SOCKET->addMethodExtended("isSecure",                  (q_method_t)SOCKET_isSecure, false, QC_RET_VALUE_ONLY, QDOM_DEFAULT, boolTypeInfo);

   // Socket::verifyPeerCertificate() returns string|nothing
   QC_SOCKET->addMethodExtended("verifyPeerCertificate",     (q_method_t)SOCKET_verifyPeerCertificate, false, QC_RET_VALUE_ONLY);

   // Socket::setCertificate(SSLCertificate $cert) returns nothing
   QC_SOCKET->addMethodExtended("setCertificate",            (q_method_t)SOCKET_setCertificate_cert, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, SSLCert->getTypeInfo(), QORE_PARAM_NO_ARG);
   // Socket::setCertificate(string $cert_pem) returns nothing
   QC_SOCKET->addMethodExtended("setCertificate",            (q_method_t)SOCKET_setCertificate_string, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, stringTypeInfo, QORE_PARAM_NO_ARG);
   // Socket::setCertificate(binary $cert_der) returns nothing
   QC_SOCKET->addMethodExtended("setCertificate",            (q_method_t)SOCKET_setCertificate_bin, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, binaryTypeInfo, QORE_PARAM_NO_ARG);

   // Socket::setPrivateKey(SSLPrivateKey $key)
   QC_SOCKET->addMethodExtended("setPrivateKey",             (q_method_t)SOCKET_setPrivateKey_key, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, SSLPrivKey->getTypeInfo(), QORE_PARAM_NO_ARG);
   // Socket::setPrivateKey(string $key_pem)
   QC_SOCKET->addMethodExtended("setPrivateKey",             (q_method_t)SOCKET_setPrivateKey_str, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, stringTypeInfo, QORE_PARAM_NO_ARG);
   // Socket::setPrivateKey(string $key_pem, string $pass)
   QC_SOCKET->addMethodExtended("setPrivateKey",             (q_method_t)SOCKET_setPrivateKey_str_str, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 2, stringTypeInfo, QORE_PARAM_NO_ARG, stringTypeInfo, QORE_PARAM_NO_ARG);
   // Socket::setPrivateKey(binary $key_der)
   QC_SOCKET->addMethodExtended("setPrivateKey",             (q_method_t)SOCKET_setPrivateKey_bin, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, binaryTypeInfo, QORE_PARAM_NO_ARG);

   QC_SOCKET->addMethodExtended("isOpen",                    (q_method_t)SOCKET_isOpen, false, QC_RET_VALUE_ONLY, QDOM_DEFAULT, boolTypeInfo);

   // Socket::setEventQueue() returns nothing
   QC_SOCKET->addMethodExtended("setEventQueue",             (q_method_t)SOCKET_setEventQueue_nothing, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo);

   // Socket::setEventQueue(Queue $queue) returns nothing
   QC_SOCKET->addMethodExtended("setEventQueue",             (q_method_t)SOCKET_setEventQueue_queue, false, QC_NO_FLAGS, QDOM_DEFAULT, nothingTypeInfo, 1, QC_QUEUE->getTypeInfo(), QORE_PARAM_NO_ARG);

   QC_SOCKET->addMethodExtended("setNoDelay",                (q_method_t)SOCKET_setNoDelay, false, QC_NO_FLAGS, QDOM_DEFAULT, bigIntTypeInfo, 1, boolTypeInfo, &True);

   QC_SOCKET->addMethodExtended("getNoDelay",                (q_method_t)SOCKET_getNoDelay, false, QC_RET_VALUE_ONLY, QDOM_DEFAULT, boolTypeInfo);

   return QC_SOCKET;
}
